/* 
 * spybmessage - dump all messages received by an application
 * Released under the MIT licence,
 * (c) 2004, François Revol, <revol@free.fr
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <OS.h>
#include <image.h>
#include <Application.h>
#include <StorageDefs.h>
#include <MessageFilter.h>
#include <MessageRunner.h>
#include <List.h>
#include <View.h>
#include <Window.h>

#define FILTER_MAX_WHAT 50

extern char **original_argv;
extern char **argv_save;
extern char **__libc_argv;
extern int original_argc;
extern int __libc_argc;

char gExePath[B_PATH_NAME_LENGTH];
image_id gExeImg = -1;
int (*gExeMainProc)(int argc, char **argv);

thread_id gWaitForBAppThID = -1;
sem_id gStdoutLock;

BMessageFilter *gAppFilter;
bool quitting = false;

/* options */
bool oAnsiOff = false; /* no ansi colors */
int oFilter = 0; /* filter output 0 = no, n = entries in FilterTab */
int oFilterOut = 0; /* filter out output 0 = no, n = entries in FilterOutTab */
uint32 gFilterTab[FILTER_MAX_WHAT];
uint32 gFilterOutTab[FILTER_MAX_WHAT];

filter_result bapp_filter(BMessage *message, 
			  BHandler **target, 
			  BMessageFilter *filter)
{
  int i;
  bool keep = false;
  if (message->what == 'plop') /* our doesn't count */
    return B_DISPATCH_MESSAGE;
  for (i = 0; i < oFilter; i++)
    if (message->what == gFilterTab[i])
      keep = true;
  if (oFilter && !keep)
    return B_DISPATCH_MESSAGE;
  for (keep = true, i = 0; i < oFilterOut; i++)
    if (message->what == gFilterOutTab[i])
      keep = false;
  if (oFilterOut && !keep)
    return B_DISPATCH_MESSAGE;
  acquire_sem(gStdoutLock);
  fprintf(stdout, "%sMessage for BApplication:%s\n",
	  oAnsiOff?"":"\033[31m", oAnsiOff?"":"\033[0m");
  message->PrintToStream();
  release_sem(gStdoutLock);
  return B_DISPATCH_MESSAGE;
}

filter_result bwin_filter(BMessage *message, 
			  BHandler **target, 
			  BMessageFilter *filter)
{
  BWindow *win;
  BView *view;
  BHandler *hand = NULL;
  int i;
  bool keep = false;
  for (i = 0; i < oFilter; i++)
    if (message->what == gFilterTab[i])
      keep = true;
  if (oFilter && !keep)
    return B_DISPATCH_MESSAGE;
  for (keep = true, i = 0; i < oFilterOut; i++)
    if (message->what == gFilterOutTab[i])
      keep = false;
  if (oFilterOut && !keep)
    return B_DISPATCH_MESSAGE;
  if (target)
    hand = *target;
  win = dynamic_cast<BWindow *>(filter->Looper());
  view = dynamic_cast<BView *>(hand);
  acquire_sem(gStdoutLock);
  fprintf(stdout, 
	  "%sMessage for %s \"%s\" of Window \"%s\":%s\n", 
	  oAnsiOff?"":"\033[31m",
	  view?"View":"Handler",
          hand?hand->Name():NULL,
	  win?win->Title():NULL,
	  oAnsiOff?"":"\033[0m");
  message->PrintToStream();
  release_sem(gStdoutLock);
  return B_DISPATCH_MESSAGE;
}

class MyHandler : public BHandler {
public:
  MyHandler();
  ~MyHandler();
  void MessageReceived(BMessage *msg);
private:
  BList fWindowList;
};

MyHandler::MyHandler()
  :BHandler("spying handler")
{
  fWindowList.MakeEmpty();
}

MyHandler::~MyHandler()
{
}

void MyHandler::MessageReceived(BMessage *msg)
{
  int i;
  BMessageFilter *afilter;
  switch (msg->what) {
  case 'plop':
    i = be_app->CountWindows();
    for (; i; i--) {
      BWindow *win = be_app->WindowAt(i-1);
      if (win && !fWindowList.HasItem(win)) {
	fWindowList.AddItem(win);
	afilter = new BMessageFilter(B_ANY_DELIVERY, 
				     B_ANY_SOURCE, 
				     bwin_filter);
	win->Lock();
	win->AddCommonFilter(afilter);
	win->Unlock();
      }
    }
    break;
  }
  BHandler::MessageReceived(msg);
}

int32 wait_for_loopers(void *arg)
{
  MyHandler *myh;
  /* wait for BApplication */
  while (!be_app)
    snooze(50000);
  gAppFilter = new BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE, bapp_filter);
  myh = new MyHandler;
  be_app->Lock();
  be_app->AddCommonFilter(gAppFilter);
  be_app->AddHandler(myh);
  be_app->Unlock();
  new BMessageRunner(BMessenger(myh), new BMessage('plop'), 100000);
  
  return 0;
}

static int usage(char *argv0)
{
  printf("usage:\n");
  printf("%s [-ch] [-f|F what] app [args...]\n", argv0);
  printf("possible options are:\n");
  printf("c\tturn off ANSI colors in the output (for logging).\n");
  printf("f\tfilter output, keep 'what' (what is either 4 letters or hex 0x...; can filter up to %d codes)\n", 
         FILTER_MAX_WHAT);
  printf("F\tfilter out output, discard 'what' (what is either 4 letters or hex 0x...; can filter up to %d codes)\n", 
         FILTER_MAX_WHAT);
  printf("h\tdisplay this help.\n");
  return 0;
}

int main(int argc, char **argv)
{
  int i;
  status_t err;
  char *trapp_name;
  if (argc < 2)
    return usage(argv[0]);
  trapp_name = argv[0];
  
  for (i = 1; i < argc; i++) {
    char filt = 0;
    if (strncmp(argv[i], "-", 1))
      break;
    if (strchr(argv[i], 'c'))
      oAnsiOff = true;
    if (strchr(argv[i], 'f'))
      filt = 'f';
    if (strchr(argv[i], 'F'))
      filt = 'F';
    if (filt) {
      uint32 what = 0;
      unsigned char c;
      i++;
      if (i >= argc) {
        printf("-%c requires an arg.\n", filt);
        return usage(argv[0]);
      }
      if (oFilter >= FILTER_MAX_WHAT) {
        printf("too many -%c\n", filt);
        return 1;
      }
      if (!strncmp(argv[i], "0x", 2)) { /* hex */
        char *end;
        what = strtoul(argv[i], &end, 16);
      } else if (strlen(argv[i]) <= 4) { /* 'what' */
        int j;
        for (j = 0; j < 4; j++) {
          c = argv[i][j];
          what = (what << 8) | c;
        }
      } else {
        printf("bad what code: %s\n", argv[i]);
        return 1;
      }
      if (filt == 'f')
        gFilterTab[oFilter++] = what;
      else
        gFilterOutTab[oFilterOut++] = what;
    }
    if (strchr(argv[i], 'h'))
      return usage(argv[0]);
  }
  if (argc - i < 1)
    return usage(argv[0]);
  argv += i;
  argc -= i;
	
  /*for (i = 0; i < argc; i++)
    printf("argv[%d] = %s\n", i, argv[i]);*/
  
  gExePath[0] = '\0';
  if (strncmp(argv[0], "/", 1)) {
    getcwd(gExePath, B_PATH_NAME_LENGTH-10);
    strcat(gExePath, "/");
  }
  strncat(gExePath, argv[0], B_PATH_NAME_LENGTH-1-strlen(gExePath));
  //printf("cmd = %s\n", gExePath);
  gExeImg = load_add_on(gExePath);
  if (gExeImg < B_OK) {
    fprintf(stderr, "load_add_on: %s\n", strerror(gExeImg));
    return 1;
  }
	
  // original are static...
  //printf("original: %d; %s\n", original_argc, *original_argv);
  //fprintf(stderr, "libc: %d; %s\n", __libc_argc, *__libc_argv);
  //fprintf(stderr, "save: %s\n", *argv_save);

  /* fixup argv caches in libroot */
  __libc_argv = argv;
  __libc_argc = argc;
  argv_save = argv;
	
  gStdoutLock = create_sem(1, "spybmsg_stdout_lock");

  err = get_image_symbol(gExeImg, "main", B_SYMBOL_TYPE_TEXT, (void **)&gExeMainProc);
  if (err < B_OK) {
    fprintf(stderr, "get_image_symbol(main): %s\n", strerror(gExeImg));
    return 1;
  }
  //printf("main @ %p\n", gExeMainProc);
  
  resume_thread(spawn_thread(wait_for_loopers, 
			     "waiting for BLoopers", 
			     B_NORMAL_PRIORITY, NULL));
  
  i = gExeMainProc(argc, argv);
  
  return i;
}

